import sys
import numpy as np
from numpy import sqrt, exp, sin, cos, tan, cosh, sinh
import numpy.linalg as LA 
import csv
from matplotlib import pyplot as plt

"""
1D band calculation by plain wave basis set
"""

#===================================
# physical constants
#===================================
pi   = 3.14159265358979323846
pi2  = 2.0 * pi
h    = 6.6260755e-34    # Js";
hbar = 1.05459e-34      # "Js";
c    = 2.99792458e8     # m/s";
e    = 1.60218e-19      # C";
e0   = 8.854418782e-12; # C<sup>2</sup>N<sup>-1</sup>m<sup>-2</sup>";
kB   = 1.380658e-23     # JK<sup>-1</sup>";
me   = 9.1093897e-31    # kg";
R    = 8.314462618      # J/K/mol
a0   = 5.29177e-11      # m";


#========================
# global configuration
#========================
mode = 'graph'   # graph|band|wf

#========================
# Crystal definition
#========================
a  = 10.0      # A, lattice parameter

#========================
# Potential
#========================
bwidth = 1.0 # A,  barrier width
bpot   = 5.0 # eV, barrier height

#========================
# Graph view
#========================
kg     = 0.0  # k point to be plotted
Emin   = 0.0
Emax   = 4.5
nE     = 21

#========================
# Band
#========================
kmin = -0.5  # in pi/a
kmax =  0.5  # in pi/a
nk = 21

# E search parameters
nEsearch = 1001
nMaxLevel = 5

#===================================
# figure configuration
#===================================
fontsize        = 12
legend_fontsize = 8


#========================
# Wave function
#========================
xwmin  = 0.0      # A
xwmax  = 3.0 * a  # A
nxw    = 101
kw     = 0.0
iLevel = 0


#==============================================
# fundamental functions
#==============================================
# 実数値に変換できない文字列をfloat()で変換するとエラーになってプログラムが終了する
# この関数は、変換できなかったらNoneを返すが、プログラムは終了させない
def pfloat(str):
    try:
        return float(str)
    except:
        return None

# pfloat()のint版
def pint(str):
    try:
        return int(str)
    except:
        return None

# 起動時引数を取得するsys.argリスト変数は、範囲外のindexを渡すとエラーになってプログラムが終了する
# egtarg()では、範囲外のindexを渡したときは、defvalを返す
def getarg(position, defval = None):
    try:
        return sys.argv[position]
    except:
        return defval

# 起動時引数を実数に変換して返す
def getfloatarg(position, defval = None):
    return pfloat(getarg(position, defval))

# 起動時引数を整数値に変換して返す
def getintarg(position, defval = None):
    return pint(getarg(position, defval))

def usage():
    print("")
    print("Usage: Variables in () are optional")
    print("  python {}".format(sys.argv[0]))
    print("  python {} (graph a bwidth bpot k Emin Emax nE)".format(sys.argv[0]))
    print("  python {} (band a bwidth bpot nG kmin kmax nk)".format(sys.argv[0]))
    print("  python {} (wf a bwidth bpot kw iLevel xwmin xwmax nxw)".format(sys.argv[0]))
    print("    pottype: rect|gauss")
    print("     ex: python {} {} {} {} {} {} {} {} {}"
            .format(sys.argv[0], 'graph', a, bwidth, bpot, kg, Emin, Emax, nE))
    print("     ex: python {} {} {} {} {} {} {} {}"
            .format(sys.argv[0], 'band', a, bwidth, bpot, kmin, kmax, nk))
    print("     ex: python {} {} {} {} {} {} {} {} {} {}"
            .format(sys.argv[0], 'wf', a, bwidth, bpot, kw, iLevel, xwmin, xwmax, nxw))

def terminate(message = None):
    print("")
    if message is not None:
        print("")
        print(message)
        print("")

    usage()
    print("")
    exit()

#==============================================
# update default values by startup arguments
#==============================================
argv = sys.argv
#if len(argv) == 1:
#    terminate()

mode    = getarg     (1, mode)
a       = getfloatarg(2, a)
bwidth  = getfloatarg(3, bwidth)
bpot    = getfloatarg(4, bpot)

if mode == 'graph':
    kg      = getfloatarg(5, kg)
    Emin    = getfloatarg(6, Emin)
    Emax    = getfloatarg(7, Emax)
    nE      = getintarg  (8, nE)
elif mode == 'band':
    kmin = getfloatarg( 5, kmin)
    kmax = getfloatarg( 6, kmax)
    nk   = getintarg  ( 7, nk)
elif mode == 'wf':
    kw     = getfloatarg( 5, kw)
    iLevel = getintarg  ( 6, iLevel)
    xwmin  = getfloatarg( 7, xwmin)
    xwmax  = getfloatarg( 8, xwmax)
    nxw    = getintarg  (9, nxw)


def reduce(x, x0):
    n = int(x / x0)
    if x < 0.0:
        n += 1
    return x - x0 * n

# rectangular barrier potential
def pot(x):
    global a
    global pottype, bwidth, bpot
    
    xred = reduce(x, a)

    if 0.0 <= xred <= bwidth:
        return bpot
    return 0.0

def build_potential(xmin, xstep, n):
    xpot = np.empty(n)
    ypot = np.empty(n)
    for i in range(n):
        xx = xmin + i * xstep
        xpot[i] = xx
        ypot[i] = pot(xx)
    return xpot, ypot

def find_iG(dij, Glist):
    for iG in range(len(Vft)):
        if Glist[iG] == dij:
            return iG
    return None

def find_Vft(dij, Glist, Vft):
    for iG in range(len(Vft)):
        if Glist[iG] == dij:
            return Vft[iG]
    return 0.0


def wf():
    global mode
    global a
    global bwidth, bpot
    global nEsearch, nMaxLevel
    global kw, iLevel
    global xwmin, xwmax, nxw

    xwstep = (xwmax - xwmin) / (nxw - 1)
    Estep = bpot / (nEsearch - 1)

    print("")
    print("=== Input parameterss ===")
    print("mode:", mode)
    print("a=", a, "A")
    print("Wave function to be plotted: k = {}  iLevel = {}".format(kw, iLevel))
    print("x range: {} - {} at {} step, {} points".format(xwmin, xwmax, xwstep, nxw))
    print("potential: w={} A  h={} eV".format(bwidth, bpot))

    print("")
    V0 = bpot
    b  = bwidth
    w  = a - b

    k = kw
    print("")
    print("at k={:8.4g}".format(k))

    d0 = None
    iband = 0
    Elist = []
    Alist = []
    for iE in range(nEsearch):
        E = 0.0 + iE * Estep
        if E == 0.0:
            continue
        if V0 <= E:
           break

        alpha = sqrt(2.0 * me * E * e) / hbar
        beta  = sqrt(2.0 * me * (V0 - E) * e) / hbar
        ka = k * pi2
        lambda_ = exp(1.0j * ka)
        alphaw = alpha * w * 1.0e-10
        betab  = beta  * b * 1.0e-10

        delta = 2.0 * cos(ka) \
              - (beta*beta - alpha*alpha)/alpha/beta * sin(alphaw) * sin(betab) \
              + 2.0 * cos(alphaw) * cosh(betab)
#        print("E=", E, "  ka=", ka, alphaw, betab, delta)
        if d0 is None:
           d0 = delta
           continue
        if d0 * delta < 0.0:
            print("E[{}]={:10.4g} eV  delta={:8.4g}".format(iband, E, delta))
            Elist.append(E - 0.5 * Estep)

            alpha *= 1.0e-10
            beta  *= 1.0e-10
            Mij = np.empty([4, 4], dtype = complex)
            Vi  = np.zeros([4, 1], dtype = complex)
            Mij[0][0] = Mij[0][1] =  1.0
            Mij[0][2] = Mij[0][3] = -1.0
            Mij[1][0] =  1.0j * alpha
            Mij[1][1] = -1.0j * alpha
            Mij[1][2] = -beta
            Mij[1][3] =  beta
            Mij[2][0] =  exp(1.0j * alphaw)
            Mij[2][1] = -exp(1.0j * alphaw)
            Mij[2][2] = -lambda_ * exp(-1.0j * betab)
            Mij[2][3] = -beta * lambda_ * exp(-1.0j * betab)
            Mij[3][0] =  1.0j * alpha * exp(-1.0j * alphaw)
            Mij[3][1] = -1.0j * alpha * exp(-1.0j * alphaw)
            Mij[3][2] = -beta * lambda_ * exp(-betab)
            Mij[3][3] =  beta * lambda_ * exp(-betab)
            Ai = LA.solve(Mij, Vi)
            Alist.append(Ai)
            print("  Ai=", Ai)

            iband += 1
            d0 = delta

    xplot, yplot = build_potential(xwmin, xwstep, nxw)

    print("")
    print("=== Calculate wave function ===")
    print("Energy levels:", Elist, "eV")
    print("at k = {}".format(k))
    print("{}-th energy level".format(iLevel))
    print("   E = {:12.6g} eV".format(Elist[iLevel]))
    ci = Alist[iLevel]
    ciLevel = [ci[i][iLevel] for i in range(len(ci))]
#    print("Wave function coefficinets")
#    for i in range(nG):
#        print("  at iG={:3d}".format(Glist[i]), "  ci={:8.4f}".format(ci[i][iLevel]))

#    print("")
#    ci2sum = sum([c * c.conjugate() for c in ci[0]]).real
#    print("  *normalization check: sum(|ci|^2)=", ci2sum, end = '')
#    ci /= sqrt(ci2sum)
#    ci2sum = sum([c * c.conjugate() for c in ci[0]]).real
#    print("     => sum(|ci|^2)=", ci2sum)

#    ywf = np.empty(nxw, dtype = complex)
#    for i in range(nxw):
#        x = xwmin + i * xwstep
#        f = 0.0
#        for iG in range(nG):
#            f += ci[iG][iLevel] * exp(1.0j * (kw+iG) * pi2/a * x)
##        print("f({})={}".format(x, f))
#        ywf[i] = f

#    charge = [f0 * f0.conjugate() for f0 in ywf]
    
    fig = plt.figure(figsize = (8, 8))
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)

    ax1.set_xlim([xwmin, xwmax])
    ax1.plot(xplot, yplot, linewidth = 0.5)
    ax1.plot(ax1.get_xlim(), [0.0, 0.0], color = 'r', linestyle = 'dashed', linewidth = 0.5)
#    ax2.set_xlim([xwmin, xwmax])
#    ax2.plot(xplot,  ywf.real, linewidth = 0.5, label = "real")
#    ax2.plot(xplot,  ywf.imag, linewidth = 0.5, label = "imaginary")
#    ax2.plot(xplot,  charge, linewidth = 1.5, label = "charge")
#    ax2.plot(ax1.get_xlim(), [0.0, 0.0], color = 'r', linestyle = 'dashed', linewidth = 0.5)
    ax1.set_xlabel("x (A)", fontsize = fontsize)
    ax1.set_ylabel("pot (x)", fontsize = fontsize)
#    ax2.legend(fontsize = legend_fontsize)
    ax1.tick_params(labelsize = fontsize)
#    ax2.tick_params(labelsize = fontsize)
#    plt.tight_layout()

    plt.pause(0.1)
    print("Press ENTER to exit>>", end = '')
    input()

    terminate()

def band():
    global mode
    global a
    global bwidth, bpot
    global kmin, kmax, nk
    global nEsearch, nMaxLevel

    kstep = (kmax - kmin) / (nk - 1)

    print("")
    print("=== Input parameterss ===")
    print("mode:", mode)
    print("a=", a, "A")
    print("potential: w={} A  h={} eV".format(bwidth, bpot))
    print("k range: {} - {} at {} step, {} points".format(kmin, kmax, kstep, nk))
    print("")

    Estep = bpot / (nEsearch - 1)

    print("")
    V0 = bpot
    b  = bwidth
    w  = a - b
    xk = [kmin + i * kstep for i in range(nk)]
    yE = np.zeros([nMaxLevel, nk])
    nMaxBand = 0
    for ik in range(nk):
        k = kmin + ik * kstep
        print("at k={:8.4g}".format(k))

        d0 = None
        iband = 0
        Elist = []
        for iE in range(nEsearch):
            E = 0.0 + iE * Estep
            if E == 0.0:
                continue
            if V0 <= E:
                break

            alpha = sqrt(2.0 * me * E * e) / hbar
            beta  = sqrt(2.0 * me * (V0 - E) * e) / hbar
            ka = k * pi2
            alphaw = alpha * w * 1.0e-10
            betab  = beta  * b * 1.0e-10
            delta = 2.0 * cos(ka) \
                - (beta*beta - alpha*alpha)/alpha/beta * sin(alphaw) * sin(betab) \
                + 2.0 * cos(alphaw) * cosh(betab)
#            print("E=", E, "  ka=", ka, alphaw, betab, delta)
            if d0 is None:
                d0 = delta
                continue
            if d0 * delta < 0.0:
                if iband < nMaxLevel:
                    print("  E[{}]={:10.4g} eV  delta={:8.4g}".format(iband, E, delta))
                    yE[iband][ik] = E - 0.5 * Estep
                    iband += 1
                    if iband > nMaxBand:
                        nMaxBand = iband
                d0 = delta

    fig = plt.figure(figsize = (8, 8))
    ax1 = fig.add_subplot(1, 1, 1)

    for iL in range(nMaxBand):
         ax1.plot(xk, yE[iL], linestyle = '', marker = 'o', markersize = 2.0,
                        markerfacecolor = 'none', markeredgecolor = 'black', markeredgewidth = 0.5)
    ax1.set_xlabel("k ($\pi$/a)", fontsize = fontsize)
    ax1.set_ylabel("E (eV)", fontsize = fontsize)
    ax1.set_xlim([-0.5, 0.5])
    ax1.set_ylim([0.0, ax1.get_ylim()[1]])
    ax1.legend(fontsize = legend_fontsize)
#    plt.tight_layout()

    plt.pause(0.1)
    print("Press ENTER to exit>>", end = '')
    input()
   
    terminate()

def graphview():
    global mode
    global a
    global bwidth, bpot
    global Emin, Emax, nE

    Estep = (Emax - Emin) / (nE - 1)

    print("")
    print("=== Input parameterss ===")
    print("mode:", mode)
    print("a=", a, "A")
    print("potential: w={} A  h={} eV".format(bwidth, bpot))
    print("Energy range: {} - {}, {} eV step  {} points".format(Emin, Emax, Estep, nE))
    print("at k = {}".format(kg))
    print("")

    V0 = bpot
    b  = bwidth
    w  = a - b
    xE = []
    yD = []
    for i in range(1, nE):
        E = Emin + i * Estep
        if V0 <= E:
            break
        alpha = sqrt(2.0 * me * E * e) / hbar
        beta  = sqrt(2.0 * me * (V0 - E) * e) / hbar
        ka = kg * pi2
        alphaw = alpha * w * 1.0e-10
        betab  = beta  * b * 1.0e-10
        delta = 2.0 * cos(ka) \
            - (beta*beta - alpha*alpha)/alpha/beta * sin(alphaw) * sin(betab) \
            + 2.0 * cos(alphaw) * cosh(betab)
#        print("a=", E, ka, alphaw, betab, delta)
        xE.append(E)
        yD.append(delta)

    fig = plt.figure(figsize = (8, 8))
    ax1 = fig.add_subplot(1, 1, 1)

    ax1.plot(xE, yD)
    ax1.set_xlim([Emin, Emax])
    ax1.plot([Emin, Emax], [0.0, 0.0], linestyle = 'dashed', color = 'r', linewidth = 0.5)
    ax1.set_xlabel("E (eV)", fontsize = fontsize)
    ax1.set_ylabel("delta", fontsize = fontsize)
#    ax1.legend(fontsize = legend_fontsize)
    ax1.tick_params(labelsize = fontsize)
    plt.tight_layout()

    plt.pause(0.1)
    print("Press ENTER to exit>>", end = '')
    input()

    terminate()

def main():
    global mode
    
    if mode == 'graph':
        graphview()
    elif mode == 'band':
        band()
    elif mode == 'wf':
        wf()
    else:
        terminate("Error: Invalid mode [{}]".format(mode))


if __name__ == "__main__":
    main()
